@@ -8,12 +8,20 @@ from .exceptions import JSONPathIndexError
 from .exceptions import JSONPathNameError
 from .exceptions import JSONPathSyntaxError
 from .exceptions import JSONPathTypeError
+from .exceptions import JSONPointerError
+from .exceptions import JSONPointerIndexError
+from .exceptions import JSONPointerKeyError
+from .exceptions import JSONPointerResolutionError
+from .exceptions import JSONPointerTypeError
 from .filter import UNDEFINED
 from .lex import Lexer
 from .match import JSONPathMatch
 from .parse import Parser
+from .patch import JSONPatch
 from .path import CompoundJSONPath
 from .path import JSONPath
+from .pointer import JSONPointer
+from .pointer import resolve
 
 __all__ = (
     "compile",
@@ -22,6 +30,7 @@ __all__ = (
     "findall",
     "finditer_async",
     "finditer",
+    "JSONPatch",
     "JSONPath",
     "JSONPathEnvironment",
     "JSONPathError",
@@ -30,16 +39,25 @@ __all__ = (
     "JSONPathNameError",
     "JSONPathSyntaxError",
     "JSONPathTypeError",
+    "JSONPointer",
+    "JSONPointerError",
+    "JSONPointerIndexError",
+    "JSONPointerKeyError",
+    "JSONPointerResolutionError",
+    "JSONPointerTypeError",
     "Lexer",
+    "match",
     "Parser",
+    "resolve",
     "UNDEFINED",
 )
 
 
 # For convenience
 DEFAULT_ENV = JSONPathEnvironment()
-findall = DEFAULT_ENV.findall
-finditer = DEFAULT_ENV.finditer
-findall_async = DEFAULT_ENV.findall_async
-finditer_async = DEFAULT_ENV.finditer_async
 compile = DEFAULT_ENV.compile  # noqa: A001
+findall = DEFAULT_ENV.findall
+findall_async = DEFAULT_ENV.findall_async
+finditer = DEFAULT_ENV.finditer
+finditer_async = DEFAULT_ENV.finditer_async
+match = DEFAULT_ENV.match
