@@ -14,7 +14,6 @@ from typing import List
 from typing import Mapping
 from typing import Optional
 from typing import Sequence
-from typing import TextIO
 from typing import Type
 from typing import Union
 
@@ -24,6 +23,8 @@ from .exceptions import JSONPathSyntaxError
 from .filter import UNDEFINED
 from .function_extensions import validate
 from .lex import Lexer
+from .match import JSONPathMatch
+from .match import NodeList
 from .parse import Parser
 from .path import CompoundJSONPath
 from .path import JSONPath
@@ -34,8 +35,9 @@ from .token import TOKEN_UNION
 from .token import Token
 
 if TYPE_CHECKING:
+    from io import IOBase
+
     from .match import FilterContextVars
-    from .match import JSONPathMatch
 
 
 class JSONPathEnvironment:
@@ -55,7 +57,7 @@ class JSONPathEnvironment:
     and overriding class attributes and/or methods. Some of these
     customizations include:
 
-    - Changing the root (`$`), self (`@`) or filter context (`#`) token with
+    - Changing the root (`$`), self (`@`) or filter context (`_`) token with
       class attributes `root_token`, `self_token` and `filter_context_token`.
     - Registering a custom lexer or parser with the class attributes
       `lexer_class` or `parser_class`. `lexer_class` must be a subclass of
@@ -67,40 +69,53 @@ class JSONPathEnvironment:
 
     ## Class attributes
 
+    Arguments:
+        filter_caching (bool): If `True`, filter expressions will be cached where
+            possible.
+
     Attributes:
-        filter_context_token: The pattern used to select extra filter context data.
-            Defaults to `"#"`.
-        intersection_token: The pattern used as the intersection operator. Defaults
-            to `"$"`.
+        filter_context_token (str): The pattern used to select extra filter context
+            data. Defaults to `"_"`.
+        intersection_token (str): The pattern used as the intersection operator.
+            Defaults to `"&"`.
+        key_token (str): The pattern used to identify the current key or index when
+            filtering a, mapping or sequence. Defaults to `"#"`.
+        keys_selector_token (str): The pattern used as the "keys" selector. Defaults to
+            `"~"`.
         lexer_class: The lexer to use when tokenizing path strings.
-        max_int_index: The maximum integer allowed when selecting array items by index.
-            Defaults to `(2**53) - 1`.
-        min_int_index: The minimum integer allowed when selecting array items by index.
-            Defaults to `-(2**53) + 1`.
+        max_int_index (int): The maximum integer allowed when selecting array items by
+            index. Defaults to `(2**53) - 1`.
+        min_int_index (int): The minimum integer allowed when selecting array items by
+            index. Defaults to `-(2**53) + 1`.
         parser_class: The parser to use when parsing tokens from the lexer.
-        root_token: The pattern used to select the root node in a JSON document.
+        root_token (str): The pattern used to select the root node in a JSON document.
             Defaults to `"$"`.
-        self_token: The pattern used to select the current node in a JSON document.
-            Defaults to `"@"`
-        union_token: The pattern used as the union operator. Defaults to `"|"`.
+        self_token (str): The pattern used to select the current node in a JSON
+            document. Defaults to `"@"`
+        union_token (str): The pattern used as the union operator. Defaults to `"|"`.
     """
 
     # These should be unescaped strings. `re.escape` will be called
     # on them automatically when compiling lexer rules.
-    intersection_token: str = "&"
-    root_token: str = "$"
-    self_token: str = "@"
-    union_token: str = "|"
-    filter_context_token: str = "#"
+    filter_context_token = "_"
+    intersection_token = "&"
+    key_token = "#"
+    keys_selector_token = "~"
+    root_token = "$"
+    self_token = "@"
+    union_token = "|"
 
-    max_int_index: int = (2**53) - 1
-    min_int_index: int = -(2**53) + 1
+    max_int_index = (2**53) - 1
+    min_int_index = -(2**53) + 1
 
     # Override these to customize path tokenization and parsing.
     lexer_class: Type[Lexer] = Lexer
     parser_class: Type[Parser] = Parser
+    match_class: Type[JSONPathMatch] = JSONPathMatch
+
+    def __init__(self, *, filter_caching: bool = True) -> None:
+        self.filter_caching = filter_caching
 
-    def __init__(self) -> None:  # noqa: D107
         self.lexer: Lexer = self.lexer_class(env=self)
         """The lexer bound to this environment."""
 
@@ -167,7 +182,7 @@ class JSONPathEnvironment:
     def findall(
         self,
         path: str,
-        data: Union[str, TextIO, Sequence[Any], Mapping[str, Any]],
+        data: Union[str, IOBase, Sequence[Any], Mapping[str, Any]],
         *,
         filter_context: Optional[FilterContextVars] = None,
     ) -> List[object]:
@@ -189,13 +204,15 @@ class JSONPathEnvironment:
 
         Raises:
             JSONPathSyntaxError: If the path is invalid.
+            JSONPathTypeError: If a filter expression attempts to use types in
+                an incompatible way.
         """
         return self.compile(path).findall(data, filter_context=filter_context)
 
     def finditer(
         self,
         path: str,
-        data: Union[str, TextIO, Sequence[Any], Mapping[str, Any]],
+        data: Union[str, IOBase, Sequence[Any], Mapping[str, Any]],
         *,
         filter_context: Optional[FilterContextVars] = None,
     ) -> Iterable[JSONPathMatch]:
@@ -214,14 +231,46 @@ class JSONPathEnvironment:
         Returns:
             An iterator yielding `JSONPathMatch` objects for each match.
 
-        Raises a `JSONPathSyntaxError` if the path is invalid.
+        Raises:
+            JSONPathSyntaxError: If the path is invalid.
+            JSONPathTypeError: If a filter expression attempts to use types in
+                an incompatible way.
         """
         return self.compile(path).finditer(data, filter_context=filter_context)
 
+    def match(
+        self,
+        path: str,
+        data: Union[str, IOBase, Sequence[Any], Mapping[str, Any]],
+        *,
+        filter_context: Optional[FilterContextVars] = None,
+    ) -> Union[JSONPathMatch, None]:
+        """Return a `JSONPathMatch` instance for the first object found in _data_.
+
+        `None` is returned if there are no matches.
+
+        Arguments:
+            path: The JSONPath as a string.
+            data: A JSON document or Python object implementing the `Sequence`
+                or `Mapping` interfaces.
+            filter_context: Arbitrary data made available to filters using
+                the _filter context_ selector.
+
+        Returns:
+            A `JSONPathMatch` object for the first match, or `None` if there were
+                no matches.
+
+        Raises:
+            JSONPathSyntaxError: If the path is invalid.
+            JSONPathTypeError: If a filter expression attempts to use types in
+                an incompatible way.
+        """
+        return self.compile(path).match(data, filter_context=filter_context)
+
     async def findall_async(
         self,
         path: str,
-        data: Union[str, TextIO, Sequence[Any], Mapping[str, Any]],
+        data: Union[str, IOBase, Sequence[Any], Mapping[str, Any]],
         *,
         filter_context: Optional[FilterContextVars] = None,
     ) -> List[object]:
@@ -233,7 +282,7 @@ class JSONPathEnvironment:
     async def finditer_async(
         self,
         path: str,
-        data: Union[str, TextIO, Sequence[Any], Mapping[str, Any]],
+        data: Union[str, IOBase, Sequence[Any], Mapping[str, Any]],
         *,
         filter_context: Optional[FilterContextVars] = None,
     ) -> AsyncIterable[JSONPathMatch]:
@@ -246,10 +295,14 @@ class JSONPathEnvironment:
         """Initialize function extensions."""
         self.function_extensions["keys"] = function_extensions.keys
         self.function_extensions["length"] = function_extensions.length
-        self.function_extensions["count"] = function_extensions.length
+        self.function_extensions["count"] = function_extensions.Count()
         self.function_extensions["match"] = function_extensions.Match()
         self.function_extensions["search"] = function_extensions.Search()
         self.function_extensions["value"] = function_extensions.value
+        self.function_extensions["isinstance"] = function_extensions.IsInstance()
+        self.function_extensions["is"] = self.function_extensions["isinstance"]
+        self.function_extensions["typeof"] = function_extensions.TypeOf()
+        self.function_extensions["type"] = self.function_extensions["typeof"]
 
     def validate_function_extension_signature(
         self, token: Token, args: List[Any]
@@ -328,6 +381,11 @@ class JSONPathEnvironment:
             `True` if the comparison between _left_ and _right_, with the
             given _operator_, is truthy. `False` otherwise.
         """
+        if isinstance(left, NodeList):
+            left = left.values_or_singular()
+        if isinstance(right, NodeList):
+            right = right.values_or_singular()
+
         if operator == "&&":
             return self.is_truthy(left) and self.is_truthy(right)
         if operator == "||":
