@@ -17,11 +17,11 @@ from typing import Union
 
 from .exceptions import JSONPathIndexError
 from .exceptions import JSONPathTypeError
-from .match import JSONPathMatch
 
 if TYPE_CHECKING:
     from .env import JSONPathEnvironment
     from .filter import BooleanExpression
+    from .match import JSONPathMatch
     from .token import Token
 
 # ruff: noqa: D102
@@ -44,7 +44,7 @@ class JSONPathSelector(ABC):
     def resolve_async(
         self, matches: AsyncIterable[JSONPathMatch]
     ) -> AsyncIterable[JSONPathMatch]:
-        """An async version of `expand`."""
+        """An async version of `resolve`."""
 
 
 class PropertySelector(JSONPathSelector):
@@ -71,7 +71,7 @@ class PropertySelector(JSONPathSelector):
                 continue
 
             with suppress(KeyError):
-                _match = JSONPathMatch(
+                _match = self.env.match_class(
                     filter_context=match.filter_context(),
                     obj=self.env.getitem(match.obj, self.name),
                     parent=match,
@@ -90,7 +90,7 @@ class PropertySelector(JSONPathSelector):
                 continue
 
             with suppress(KeyError):
-                _match = JSONPathMatch(
+                _match = self.env.match_class(
                     filter_context=match.filter_context(),
                     obj=await self.env.getitem_async(match.obj, self.name),
                     parent=match,
@@ -134,11 +134,11 @@ class IndexSelector(JSONPathSelector):
             if isinstance(match.obj, Mapping):
                 # Try the string representation of the index as a key.
                 with suppress(KeyError):
-                    _match = JSONPathMatch(
+                    _match = self.env.match_class(
                         filter_context=match.filter_context(),
                         obj=self.env.getitem(match.obj, self._as_key),
                         parent=match,
-                        parts=match.parts + (self.index,),
+                        parts=match.parts + (self._as_key,),
                         path=f"{match.path}['{self.index}']",
                         root=match.root,
                     )
@@ -147,7 +147,7 @@ class IndexSelector(JSONPathSelector):
             elif isinstance(match.obj, Sequence):
                 norm_index = self._normalized_index(match.obj)
                 with suppress(IndexError):
-                    _match = JSONPathMatch(
+                    _match = self.env.match_class(
                         filter_context=match.filter_context(),
                         obj=self.env.getitem(match.obj, self.index),
                         parent=match,
@@ -165,11 +165,11 @@ class IndexSelector(JSONPathSelector):
             if isinstance(match.obj, Mapping):
                 # Try the string representation of the index as a key.
                 with suppress(KeyError):
-                    _match = JSONPathMatch(
+                    _match = self.env.match_class(
                         filter_context=match.filter_context(),
                         obj=await self.env.getitem_async(match.obj, self._as_key),
                         parent=match,
-                        parts=match.parts + (self.index,),
+                        parts=match.parts + (self._as_key,),
                         path=f"{match.path}['{self.index}']",
                         root=match.root,
                     )
@@ -178,7 +178,7 @@ class IndexSelector(JSONPathSelector):
             elif isinstance(match.obj, Sequence):
                 norm_index = self._normalized_index(match.obj)
                 with suppress(IndexError):
-                    _match = JSONPathMatch(
+                    _match = self.env.match_class(
                         filter_context=match.filter_context(),
                         obj=await self.env.getitem_async(match.obj, self.index),
                         parent=match,
@@ -201,11 +201,11 @@ class KeysSelector(JSONPathSelector):
     def _keys(self, match: JSONPathMatch) -> Iterable[JSONPathMatch]:
         if isinstance(match.obj, Mapping):
             for i, key in enumerate(match.obj.keys()):
-                _match = JSONPathMatch(
+                _match = self.env.match_class(
                     filter_context=match.filter_context(),
                     obj=key,
                     parent=match,
-                    parts=match.parts + (self.env.keys_selector_token, i),
+                    parts=match.parts + (f"{self.env.keys_selector_token}{key}",),
                     path=f"{match.path}[{self.env.keys_selector_token}][{i}]",
                     root=match.root,
                 )
@@ -269,7 +269,7 @@ class SliceSelector(JSONPathSelector):
             step = self.slice.step or 1
             for obj in self.env.getitem(match.obj, self.slice):
                 norm_index = self._normalized_index(match.obj, idx)
-                _match = JSONPathMatch(
+                _match = self.env.match_class(
                     filter_context=match.filter_context(),
                     obj=obj,
                     parent=match,
@@ -292,7 +292,7 @@ class SliceSelector(JSONPathSelector):
             step = self.slice.step or 1
             for obj in await self.env.getitem_async(match.obj, self.slice):
                 norm_index = self._normalized_index(match.obj, idx)
-                _match = JSONPathMatch(
+                _match = self.env.match_class(
                     filter_context=match.filter_context(),
                     obj=obj,
                     parent=match,
@@ -317,7 +317,7 @@ class WildSelector(JSONPathSelector):
                 continue
             if isinstance(match.obj, Mapping):
                 for key, val in match.obj.items():
-                    _match = JSONPathMatch(
+                    _match = self.env.match_class(
                         filter_context=match.filter_context(),
                         obj=val,
                         parent=match,
@@ -329,7 +329,7 @@ class WildSelector(JSONPathSelector):
                     yield _match
             elif isinstance(match.obj, Sequence):
                 for i, val in enumerate(match.obj):
-                    _match = JSONPathMatch(
+                    _match = self.env.match_class(
                         filter_context=match.filter_context(),
                         obj=val,
                         parent=match,
@@ -346,7 +346,7 @@ class WildSelector(JSONPathSelector):
         async for match in matches:
             if isinstance(match.obj, Mapping):
                 for key, val in match.obj.items():
-                    _match = JSONPathMatch(
+                    _match = self.env.match_class(
                         filter_context=match.filter_context(),
                         obj=val,
                         parent=match,
@@ -358,7 +358,7 @@ class WildSelector(JSONPathSelector):
                     yield _match
             elif isinstance(match.obj, Sequence):
                 for i, val in enumerate(match.obj):
-                    _match = JSONPathMatch(
+                    _match = self.env.match_class(
                         filter_context=match.filter_context(),
                         obj=val,
                         parent=match,
@@ -382,7 +382,7 @@ class RecursiveDescentSelector(JSONPathSelector):
                 if isinstance(val, str):
                     pass
                 elif isinstance(val, (Mapping, Sequence)):
-                    _match = JSONPathMatch(
+                    _match = self.env.match_class(
                         filter_context=match.filter_context(),
                         obj=val,
                         parent=match,
@@ -398,7 +398,7 @@ class RecursiveDescentSelector(JSONPathSelector):
                 if isinstance(val, str):
                     pass
                 elif isinstance(val, (Mapping, Sequence)):
-                    _match = JSONPathMatch(
+                    _match = self.env.match_class(
                         filter_context=match.filter_context(),
                         obj=val,
                         parent=match,
@@ -490,7 +490,7 @@ class ListSelector(JSONPathSelector):
 class Filter(JSONPathSelector):
     """A filter selector."""
 
-    __slots__ = ("expression",)
+    __slots__ = ("expression", "cacheable_nodes")
 
     def __init__(
         self,
@@ -501,6 +501,8 @@ class Filter(JSONPathSelector):
     ) -> None:
         super().__init__(env=env, token=token)
         self.expression = expression
+        # Compile-time check for cacheable nodes.
+        self.cacheable_nodes = self.expression.cacheable_nodes()
 
     def __str__(self) -> str:
         return f"[?({self.expression})]"
@@ -508,6 +510,11 @@ class Filter(JSONPathSelector):
     def resolve(  # noqa: PLR0912
         self, matches: Iterable[JSONPathMatch]
     ) -> Iterable[JSONPathMatch]:
+        if self.cacheable_nodes and self.env.filter_caching:
+            expr = self.expression.cache_tree()
+        else:
+            expr = self.expression
+
         for match in matches:
             if isinstance(match.obj, Mapping):
                 for key, val in match.obj.items():
@@ -519,8 +526,8 @@ class Filter(JSONPathSelector):
                         current_key=key,
                     )
                     try:
-                        if self.expression.evaluate(context):
-                            _match = JSONPathMatch(
+                        if expr.evaluate(context):
+                            _match = self.env.match_class(
                                 filter_context=match.filter_context(),
                                 obj=val,
                                 parent=match,
@@ -545,8 +552,8 @@ class Filter(JSONPathSelector):
                         current_key=i,
                     )
                     try:
-                        if self.expression.evaluate(context):
-                            _match = JSONPathMatch(
+                        if expr.evaluate(context):
+                            _match = self.env.match_class(
                                 filter_context=match.filter_context(),
                                 obj=obj,
                                 parent=match,
@@ -564,6 +571,11 @@ class Filter(JSONPathSelector):
     async def resolve_async(  # noqa: PLR0912
         self, matches: AsyncIterable[JSONPathMatch]
     ) -> AsyncIterable[JSONPathMatch]:
+        if self.cacheable_nodes and self.env.filter_caching:
+            expr = self.expression.cache_tree()
+        else:
+            expr = self.expression
+
         async for match in matches:
             if isinstance(match.obj, Mapping):
                 for key, val in match.obj.items():
@@ -576,14 +588,14 @@ class Filter(JSONPathSelector):
                     )
 
                     try:
-                        result = await self.expression.evaluate_async(context)
+                        result = await expr.evaluate_async(context)
                     except JSONPathTypeError as err:
                         if not err.token:
                             err.token = self.token
                         raise
 
                     if result:
-                        _match = JSONPathMatch(
+                        _match = self.env.match_class(
                             filter_context=match.filter_context(),
                             obj=val,
                             parent=match,
@@ -605,13 +617,13 @@ class Filter(JSONPathSelector):
                     )
 
                     try:
-                        result = await self.expression.evaluate_async(context)
+                        result = await expr.evaluate_async(context)
                     except JSONPathTypeError as err:
                         if not err.token:
                             err.token = self.token
                         raise
                     if result:
-                        _match = JSONPathMatch(
+                        _match = self.env.match_class(
                             filter_context=match.filter_context(),
                             obj=obj,
                             parent=match,
@@ -626,7 +638,13 @@ class Filter(JSONPathSelector):
 class FilterContext:
     """A filter expression context."""
 
-    __slots__ = ("current", "current_key", "env", "root", "extra_context")
+    __slots__ = (
+        "current_key",
+        "current",
+        "env",
+        "extra_context",
+        "root",
+    )
 
     def __init__(
         self,
